#include <mpi.h>
#include <fftw3.h>

#include <fftk.h>


void test(string basis, string basis_option, int Nx, int Ny, int Nz, int iter);
Real f(string basis, string basis_option, int rx, int ry, int rz);
void Print_large_Fourier_elements(Array<Complex,3> A, string array_name="A");

int Get_kx(int lx);
int Get_ky(int ly);
int Get_kz(int lz);


int my_id,numprocs, Nx,Ny,Nz, fx_start,fy_start,fz_start;

string basis,basis_option;

int main(int argc, char** argv)
{

  	MPI_Init(&argc, &argv);
  	Nx=8;
  	Ny=1;
  	Nz=8;

  	test("FFF", "FFF", Nx, Ny, Nz, 10);

	MPI_Finalize();
	return 0;
}


void test(string basis1, string basis_option1, int Nx, int Ny, int Nz, int iter)
{
  	MPI_Comm_rank(MPI_COMM_WORLD, &my_id);
	MPI_Comm_size(MPI_COMM_WORLD, &numprocs);

	basis=basis1;
	basis_option=basis_option1;

	//Nx=Ny=N;
	//Nz=N;

	int rows=1;

  	if (Nx<numprocs)		//For pencil
  		rows=numprocs/Nx;

	FFTK fftk;
	fftk.Init(basis,Nx,Ny,Nz,rows);

	Array<Complex,3> A(fftk.Get_FA_shape());
	Array<Real,3> Ar(fftk.Get_RA_shape());
	Array<Real, 3> Ar_init(fftk.Get_RA_shape());

	Ar_init=0;

	int maxrx=Ar.extent(0);
	int maxry=Ar.extent(1);
	int maxrz=Ar.extent(2);


	TinyVector<int,3> FA_start = fftk.Get_FA_start();
	fx_start=FA_start(0);
	fy_start=FA_start(1);
	fz_start=FA_start(2);

	TinyVector<int,3> RA_start = fftk.Get_RA_start();
	int rx_start=RA_start(0);
	int ry_start=RA_start(1);
	int rz_start=RA_start(2);

	for (int rx=0; rx<maxrx; rx++)
		for (int ry=0; ry<maxry; ry++)
			for (int rz=0; rz<maxrz; rz++){
				Ar_init(rx, ry, rz) = f(basis, basis_option, rx_start + rx, ry_start + ry, rz);
				// Ar_init(rx, ry, rz) =  my_id*Ar_init.size() + (rx*Ar_init.extent(1)*Ar_init.extent(2)) + ry*Ar_init.extent(2) + rz;
			}

	//Zero pad
	if (basis[2] != 'S')
		fftk.Zero_pad_last_plane(Ar_init);

	// cout << Ar_init << endl;

/*	if (my_id==1)
		sleep(1);
	cout << Ar_init << endl;	*/	


	// double local_energy = sum(sqr(Ar_init));
	// double total_energy;
	// MPI_Reduce(&local_energy, &total_energy, 1, MPI_Real, MPI_SUM, 0, MPI_COMM_WORLD);

/*	if (my_id == 0)
		cout << "Before transform energy = " << total_energy << endl;*/

    Real start_time;
    Real end_time;

	Ar = Ar_init;

    Real local_time=0, local_time_max;
    MPI_Barrier(MPI_COMM_WORLD);


	time_t start;
	time_t end;


	// if (my_id==0) cout << Ar << endl;

	if (Ny>1) {
		fftk.Forward_transform(basis_option, Ar, A);

		// if (my_id==0) cout << A << endl;

		Print_large_Fourier_elements(A);
		fftk.Inverse_transform(basis_option, A, Ar);
	}
	else {
		fftk.Forward_transform(basis_option, Ar(Range::all(),0,Range::all()), A(Range::all(),0,Range::all()));

		// if (my_id==0) cout << A << endl;

		Print_large_Fourier_elements(A);
		fftk.Inverse_transform(basis_option, A(Range::all(),0,Range::all()), Ar(Range::all(),0,Range::all()));
	}	

	// Ar*=2;

	// cout << Ar << endl;

	// Ar/=2;

	// if (my_id==0) cout << Ar << endl;

	//return;

	time (&start);
	/*for (int i=0; i<iter; i++) {
		fftk.Forward_transform(basis_option, Ar, A);
		fftk.Inverse_transform(basis_option, A, Ar);
	}*/
	time(&end);

	double *timer_avg, *timer_max, *timer_min;

	timer_avg = new double[22];
	timer_max = new double[22];
	timer_min = new double[22];

	MPI_Reduce(fftk.timer, timer_avg, 22, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
	MPI_Reduce(fftk.timer, timer_max, 22, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
	MPI_Reduce(fftk.timer, timer_min, 22, MPI_DOUBLE, MPI_MIN, 0, MPI_COMM_WORLD);


	Real dif = difftime(end, start);
	if (my_id == 0) {
		cout << "Time taken = " << dif << endl;
		cout << "max(abs(Ar-Ar_init)) = " << max(abs(Ar-Ar_init)) << endl;
		for (int i=0; i<22; i++)
			printf("timer[%02d] (avg/max/min): %E %E %E \n", i, timer_avg[i]/22, timer_max[i], timer_min[i]);

	}
}

void Print_large_Fourier_elements(Array<Complex,3> A, string array_name)
{

	for (int p=0; p<numprocs; p++) {
		if (my_id==p) {
			for (int lx=0; lx<A.extent(0); lx++)
				for (int ly=0; ly<A.extent(1); ly++)
					for (int lz=0; lz<A.extent(2); lz++)
						if (abs(A(lx, ly, lz)) > 1E-6) {
							cout << "my_id = " << my_id <<  " vect(k) = (" << Get_kx(lx) << "," << Get_ky(ly) << "," << Get_kz(lz) <<");  " << array_name << "(k) = " << A(lx, ly, lz) << '\n';
						}
		}
		MPI_Barrier(MPI_COMM_WORLD);
	}
}

Real f(string basis, string basis_option, int rx, int ry, int rz)
{
	Real k0 = 1;
	Real x,y,z;
	
	Real Lx=2*M_PI;
	Real Ly=2*M_PI;
	Real Lz=2*M_PI;
	if (basis == "FFF") {
		x = rx*Lx/Nx;
		y = ry*Ly/Ny;
		z = rz*Lz/Nz;
		if (Ny>1)
			return 8*sin(k0*x)*sin(k0*y)*sin(k0*z);
		else
			return 8*sin(k0*x)*sin(k0*z);
	}
	
	else if (basis == "SFF") {
		Lx=M_PI;
		x = (rx+0.5)*Lx/Nx;
		y = ry*Ly/Ny;
		z = rz*Lz/Nz;

		if (basis_option=="SFF") {
			if (Ny>1)
				return 8*sin(k0*x)*cos(k0*y)*cos(k0*z);
			else
				return 8*sin(k0*x)*cos(k0*z);
		}
		else if (basis_option=="CFF") {
			if (Ny>1)
				return 8*cos(k0*x)*cos(k0*y)*cos(k0*z);
			else
				return 8*cos(k0*x)*cos(k0*z);
		}
	}
	
	else if (basis == "SSF") {
		x = (rx+0.5)*Lx/Nx;
		y = (ry+0.5)*Ly/Ny;
		z = rz*Lz/Nz;

		if (basis_option=="SSF") {
			if (Ny>1)
				return 8*sin(k0*x)*sin(k0*y)*cos(k0*z);
			else
				return 8*sin(k0*x)*cos(k0*z);
		}
		else if (basis_option=="CCF") {
			if (Ny>1)
				return 8*cos(k0*x)*cos(k0*y)*cos(k0*z);
			else
				return 8*cos(k0*x)*cos(k0*z);
		}
	}
	
	else if (basis == "SSS") {
		x = (rx+0.5)*Lx/Nx;
		y = (ry+0.5)*Ly/Ny;
		z = (rz+0.5)*Lz/Nz;

		if (basis_option=="SSS") {
			if (Ny>1)
				return 8*sin(k0*x)*sin(k0*y)*sin(k0*z);
			else
				return 8*sin(k0*x)*sin(k0*z);
		}
		else if (basis_option=="CCC") {
			if (Ny>1)
				return 8*cos(k0*x)*cos(k0*y)*cos(k0*z);
			else
				return 8*cos(k0*x)*cos(k0*z);
		}
	}
	
	else if (basis == "ChFF") {
		x = rx*M_PI/(Nx-1);
		y = ry*Ly/Ny;
		z = rz*Lz/Nz;
	}

	
	
//	return 4*sin(k0*x)*cos(k0*z);
//	return 4*(2*cos(k0*x)*cos(k0*x)-1)*cos(k0*y)*cos(k0*z); // Cheb
//	return 4*(cos(k0*x))*cos(k0*y)*cos(k0*z);  // Cheb
//	return 2*(cos(k0*x))*cos(k0*z);  // Cheb (101)
	// return 8*sin(k0*x)*sin(k0*y)*sin(k0*z);
	// return 4*sin(k0*x)*sin(k0*z);
}

int Get_kx(int lx) {
	if (basis[0]=='F') {
		if (lx+fx_start>Nx/2)
			return (lx+fx_start-Nx);
		else
			return (lx+fx_start);
	}
	else if (basis[0]=='S') {
			return (lx+fx_start);
	}
	return 0;
}

int Get_ky(int ly) {
	if (basis[1]=='F') {
		if (ly+fy_start>Ny/2)
			return (ly+fy_start-Ny);
		else
			return (ly+fy_start);
	}
	else if (basis[0]=='S') {
			return (ly+fy_start);
	}

	return 0;
}

int Get_kz(int lz) {
	if (basis[2]=='F') {
		if (lz+fz_start>Nz/2)
			return (lz+fz_start-Nz);
		else
			return (lz+fz_start);
	}
	else if (basis[0]=='S') {
			return (lz+fz_start);
	}

	return 0;
}
